import pytest

from pyrfc_read.utils import field_value, format_fields, format_value, format_wheres


@pytest.mark.parametrize(
    "value, field_type, expected",
    [
        ("123", "I", 123),
        ("0123", "I", 123),
        (123, "I", 123),
        ("ABC", "C", "ABC"),
        ("ABC    ", "C", "ABC"),
        (123.45, "F", 123.45),
        ("a 123.45", "F", "a 123.45"),
    ],
)
def test_field_value(value, field_type, expected):
    assert field_value(value, field_type) == expected


def test_format_fields():
    assert format_fields(None) == ""
    assert format_fields("") == ""
    assert format_fields("*") == ""
    assert format_fields("ABC") == [{"FIELDNAME": "ABC"}]
    assert format_fields(("ABC", "DEF")) == [
        {"FIELDNAME": "ABC"},
        {"FIELDNAME": "DEF"},
    ]
    assert format_fields(["ABC", "DEF"]) == [
        {"FIELDNAME": "ABC"},
        {"FIELDNAME": "DEF"},
    ]

    # Improper fields format
    fields = object()
    with pytest.raises(ValueError):
        format_fields(fields)


@pytest.mark.parametrize(
    "value, field_type, field_length, expected",
    [
        ("123", "C", 7, "'0000123'"),
        ("123.45", "C", 7, "'0123.45'"),
        ("ABC", "C", 7, "'ABC'"),
        (123, "I", 7, "'123'"),
        (123.45, "F", 7, "'123.45'"),
        (123.45, "P", 7, "'123.45'"),
    ],
)
def test_format_value(value, field_type, field_length, expected):
    assert format_value(value, field_type, field_length) == expected


def test_format_wheres():
    field_info = {
        "Field1": {"INTTYPE": "C", "LENG": 7},
        "Field2": {"INTTYPE": "I", "LENG": 2},
    }
    assert format_wheres(None, field_info) == [{"TEXT": ""}]
    assert format_wheres("", field_info) == [{"TEXT": ""}]
    assert format_wheres("ABC", field_info) == [{"TEXT": "ABC"}]
    assert format_wheres(["ABC", "DEF"], field_info) == [
        {"TEXT": "ABC AND DEF"},
    ]
    assert format_wheres(
        "A long query condition that has to be split into many lines, as it's too long for SAP.",
        field_info,
    ) == [
        {
            "TEXT": "A long query condition that has to be split into many lines, as it's too"
        },
        {"TEXT": " long for SAP."},
    ]
    assert format_wheres([["Field1", "in", ["ABC", "DEF", "GHI"]]], field_info) == [
        {"TEXT": "(Field1 = 'ABC' OR Field1 = 'DEF' OR Field1 = 'GHI')"}
    ]
    assert format_wheres([["Field1", "not in", ["ABC", "DEF", "GHI"]]], field_info) == [
        {"TEXT": "(Field1 <> 'ABC' AND Field1 <> 'DEF' AND Field1 <> 'GHI')"}
    ]
    assert format_wheres([[["Field1", "Field2"], "in", []]], field_info) == [
        {"TEXT": ""}
    ]
    assert format_wheres(
        [[["Field1", "Field2"], "in", [["ABC", 1], ["DEF", 2], ["GHI", 3]]]], field_info
    ) == [
        {
            "TEXT": "((Field1 = 'ABC' AND Field2 = '1') OR (Field1 = 'DEF' AND Field2 = '2') "
        },
        {"TEXT": "OR (Field1 = 'GHI' AND Field2 = '3'))"},
    ]
    assert format_wheres([[["Field1", "Field2"], "not in", []]], field_info) == [
        {"TEXT": ""}
    ]
    assert format_wheres(
        [[["Field1", "Field2"], "not in", [["ABC", 1], ["DEF", 2], ["GHI", 3]]]],
        field_info,
    ) == [
        {
            "TEXT": "((Field1 <> 'ABC' OR Field2 <> '1') AND (Field1 <> 'DEF' OR Field2 <> "
        },
        {"TEXT": "'2') AND (Field1 <> 'GHI' OR Field2 <> '3'))"},
    ]



import logging
import textwrap
from collections.abc import Iterable

logger = logging.getLogger(__name__)


def field_value(value, field_type: str):
    """Convert value from a field to the proper Python type, given its field_type

    Args:
        value: Value to convert
        field_type: Field type

    Returns:
        Converted value
    """

    match field_type:
        case "I":
            return int(value)
        case "F" | "P":
            try:
                return float(value)
            except ValueError:
                return value
        case "C":
            return value.rstrip()


def format_fields(fields) -> list[dict[str, str]]:
    """Format fields into SAP Query format

    Args:
        fields: String or Iterable of field(s) to format

    Returns:
        Formatted fields as list of dict['FIELDNAME', str]
    """

    if not fields or fields == "*":
        fields = ""
    elif isinstance(fields, str):
        fields = [{"FIELDNAME": fields}]
    elif isinstance(fields, Iterable):
        fields = [
            field if isinstance(field, dict) else {"FIELDNAME": field}
            for field in fields
        ]
    else:
        raise ValueError(f"Improper 'fields' format: {fields}")

    return fields


def format_value(value, field_type: str, field_length: int) -> str:
    """Format field value into SAP Query format

    Args:
        value: Value to format
        field_type: Field datatype
            (e.g. 'C' for character string, 'I' for integer, 'F' for floating point number)
        field_length: Field length

    Returns:
        Formatted value
    """

    if field_type == "C":
        try:
            try:
                value = int(value)
            except ValueError:
                value = float(value)
            # value look numeric, so left-pad it with zeroes
            value = str(value).zfill(field_length)
        except ValueError:
            # value doesn't look numeric
            pass

    # enclose in quotes
    value = f"'{value}'"

    return value


def format_wheres(
    wheres, field_info: dict[str, dict[str, str]]
) -> list[dict[str, str]]:
    """Format wheres into SAP Query format

    Args:
        wheres: Where(s) to format
        field_info: Info for each field in the table (use `field_info()` to get this)

    Returns:
        Formatted wheres as list of dict['TEXT', str]
    """

    if not wheres:
        return [{"TEXT": ""}]
    elif isinstance(wheres, str):
        options = wheres
    elif isinstance(wheres, (list, set)):
        options = ""
        for where in wheres:
            if len(options) > 0:
                options += " AND "
            if isinstance(where, str):
                options += where
            elif isinstance(where, list):
                if len(where) == 1 and isinstance(where[0], str):
                    options += where[0]
                elif len(where) == 3:
                    if all(isinstance(where[i], str) for i in range(3)):
                        options += " ".join(where)
                    elif (
                        all(isinstance(where[i], str) for i in range(2))
                        and where[1].lower() in ("in", "not in")
                        and isinstance(where[2], Iterable)
                    ):
                        # This field considers a list
                        if len(where[2]) == 0:
                            logger.warning(
                                f"Ignoring 'wheres' with empty list: {where}"
                            )
                            if len(wheres) == 1:
                                return [{"TEXT": ""}]
                            continue
                        # Find the field format
                        field = where[0]
                        field_type = field_info[field]["INTTYPE"]
                        field_length = int(field_info[field]["LENG"])
                        if where[1].lower() == "in":
                            # Field must be in list
                            for n, item in enumerate(where[2]):
                                item_str = format_value(item, field_type, field_length)
                                if n == 0:
                                    options += f"({field} = {item_str}"
                                else:
                                    options += f" OR {field} = {item_str}"
                        else:
                            # Field must not be in list
                            for n, item in enumerate(where[2]):
                                item_str = format_value(item, field_type, field_length)
                                if n == 0:
                                    options += f"({field} <> {item_str}"
                                else:
                                    options += f" AND {field} <> {item_str}"
                        options += ")"
                    elif (
                        isinstance(where[0], Iterable)
                        and all(isinstance(i, str) for i in where[0])
                        and where[1].lower() in ("in", "not in")
                        and isinstance(where[2], Iterable)
                    ):
                        # This list of fields must be within allowed combinations (a list of tuples)
                        if len(where[2]) == 0:
                            logger.warning(
                                f"Ignoring 'wheres' with empty list: {where}"
                            )
                            if len(wheres) == 1:
                                return [{"TEXT": ""}]
                            continue
                        # Find the fields format
                        fields = where[0]
                        fields_type = {
                            field: field_info[field]["INTTYPE"] for field in fields
                        }
                        fields_length = {
                            field: int(field_info[field]["LENG"]) for field in fields
                        }
                        if where[1].lower() == "in":
                            # Fields must be in list
                            for n, items in enumerate(where[2]):
                                if n == 0:
                                    options += "("
                                else:
                                    options += " OR "
                                for n_field, field in enumerate(where[0]):
                                    item = items[n_field]
                                    field_type = fields_type[field]
                                    field_length = fields_length[field]
                                    item_str = format_value(
                                        item, field_type, field_length
                                    )
                                    if n_field == 0:
                                        options += f"({field} = {item_str}"
                                    else:
                                        options += f" AND {field} = {item_str}"
                                options += ")"
                        else:
                            # Fields must not be in list
                            for n, items in enumerate(where[2]):
                                if n == 0:
                                    options += "("
                                else:
                                    options += " AND "
                                for n_field, field in enumerate(where[0]):
                                    item = items[n_field]
                                    field_type = fields_type[field]
                                    field_length = fields_length[field]
                                    item_str = format_value(
                                        item, field_type, field_length
                                    )
                                    if n_field == 0:
                                        options += f"({field} <> {item_str}"
                                    else:
                                        options += f" OR {field} <> {item_str}"
                                options += ")"
                        options += ")"
                    else:
                        raise ValueError(f"Improper 'wheres' format: {where}")
                else:
                    raise ValueError(f"Improper 'wheres' format: {where}")
            else:
                raise ValueError(f"Improper 'wheres' format: {where}")
    else:
        raise ValueError(f"Improper 'wheres' format: {wheres}")

    # Split where clauses into max 72-character lines
    wheres = [
        {"TEXT": line} for line in textwrap.wrap(options, 72, drop_whitespace=False)
    ]

    return wheres



import codecs
import os.path

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


def read_requirements(req_type=None):
    fname = f"requirements-{req_type}.txt" if req_type else "requirements.txt"
    with open(fname) as f:
        requires = (line.strip() for line in f)
        return [req for req in requires if req and not req.startswith("#")]


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="pyrfc-read",
    version=get_version("pyrfc_read/__init__.py"),
    author="Alex Marvin",
    author_email="alex.marvin@gmail.com",
    description="Read data from SAP R/3 Systems",
    url="https://github.com/amarvin/pyrfc-read",
    packages=find_packages(exclude=["tests"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=read_requirements(),
    extras_require={
        "dev": read_requirements("dev"),
        "test": read_requirements("test"),
    },
)



import logging
from collections.abc import Iterable

import pyrfc

from pyrfc_read.utils import field_value, format_fields, format_wheres

logger = logging.getLogger(__name__)

# Protect against missing SAP NW SDK prerequisite for pyrfc
try:
    BaseConnection = pyrfc.Connection
except AttributeError:
    logger.error("pyrfc not configured correctly, likely due to missing SAP NW SDK.")
    BaseConnection = object


class Connection(BaseConnection):
    def query(
        self,
        table: str,
        fields: None | list[str] = None,
        wheres=None,
        max_rows: int = 50,
        from_row: int = 0,
        field_info: None | bool | dict[str, dict[str, str]] = None,
        delimiter: str = "⁂",
        batch_rows: None | int = None,
        chunk_rows: int = 10000,
        make_unique: bool = True,
    ) -> list[list]:
        """Query table data from SAP R/3 System

        Args:
            table: Table to query
            fields: Fields to query
            wheres: Where condition(s) for query
            max_rows: Maximum number of rows to return
            from_row: Starting row
            field_info: Metadata for each field in the table (None to query field_info, False to not)
            delimiter: Delimiter character used in data transfers
            batch_rows: Number of rows to query at a time
            chunk_rows: Query data in chuncks of this size, if wheres have many entries
            make_unique: Convert wheres to unique values

        Returns:
            List of list, with data from table
        """

        # Convert all wheres to unique values
        if make_unique and wheres is not None:
            wheres_with_arrays = (
                i
                for i, where in enumerate(wheres)
                if isinstance(where, list)
                if len(where) == 3
                if isinstance(where[0], str)
                if where[1].lower() in ("in", "not in")
                if not isinstance(where[2], str)
                if isinstance(where[2], Iterable)
            )
            for i in wheres_with_arrays:
                # Reduce array to unique values
                uniques = set()
                wheres[i][2] = [
                    x
                    for x in wheres[i][2]
                    if x not in uniques and (uniques.add(x) or True)
                ]
            return self.query(
                table,
                fields,
                wheres,
                max_rows,
                from_row,
                field_info,
                delimiter,
                batch_rows,
                chunk_rows,
                make_unique=False,
            )

        # Get field metadata
        if field_info is None:
            field_info = self.field_info(table)

        # Convert inputs to SAP Query format
        fields = format_fields(fields)

        # Detect if chinking wheres is needed
        chunk_field = None
        if wheres is not None:
            for where in wheres:
                if (
                    isinstance(where, list)
                    and len(where) == 3
                    and isinstance(where[0], str)
                    and where[1].lower() == "in"
                    and not isinstance(where[2], str)
                    and isinstance(where[2], Iterable)
                ):
                    chunk_list = where[2]
                    if len(chunk_list) > chunk_rows:
                        # This field needs chunking
                        chunk_field = where[0]

                        # Remove this where
                        wheres.remove(where)

                        break

        # Query data
        if not chunk_field:
            if not batch_rows:
                data = self._query(
                    table, fields, wheres, max_rows, from_row, field_info, delimiter
                )
            else:
                # Batch queries
                data = []
                max_rows_batch = min(max_rows, batch_rows) if max_rows else batch_rows
                last_data = self._query(
                    table,
                    fields,
                    wheres,
                    max_rows_batch,
                    from_row,
                    field_info,
                    delimiter,
                )
                data.extend(last_data)

                # Query more data, if needed
                last_row = max_rows_batch + from_row
                while len(last_data) == max_rows_batch and (
                    not max_rows or last_row < max_rows
                ):
                    # Query more data
                    last_data = self._query(
                        table,
                        fields,
                        wheres,
                        max_rows_batch,
                        last_row,
                        field_info,
                        delimiter,
                    )
                    data.extend(last_data)
                    last_row += max_rows_batch
        else:
            # Chunk queries
            data = []
            for i in range(0, len(chunk_list), chunk_rows):
                chunk = chunk_list[i : i + chunk_rows]
                new_wheres = wheres + [[chunk_field, "in", chunk]]
                if not batch_rows:
                    last_data = self.query(
                        table,
                        fields,
                        new_wheres,
                        max_rows,
                        from_row,
                        field_info,
                        delimiter,
                        batch_rows,
                        chunk_rows,
                    )
                    data.extend(last_data)
                else:
                    # Batch queries
                    max_rows_batch = (
                        min(max_rows, batch_rows) if max_rows else batch_rows
                    )
                    last_data = self.query(
                        table,
                        fields,
                        new_wheres,
                        max_rows_batch,
                        from_row,
                        field_info,
                        delimiter,
                        batch_rows,
                        chunk_rows,
                    )
                    data.extend(last_data)

                    # Query more data, if needed
                    last_row = max_rows_batch + from_row
                    while len(last_data) == max_rows_batch and (
                        not max_rows or last_row < max_rows
                    ):
                        # Query more data
                        last_data = self.query(
                            table,
                            fields,
                            new_wheres,
                            max_rows_batch,
                            last_row,
                            field_info,
                            delimiter,
                            batch_rows,
                            chunk_rows,
                        )
                        data.extend(last_data)
                        last_row += max_rows_batch

        # Return results
        return data

    def _query(
        self,
        table: str,
        fields: None | list[str],
        wheres,
        max_rows: int,
        from_row: int,
        field_info: None | dict[str, dict[str, str]],
        delimiter: str,
    ) -> list[list]:
        """Low-level query function

        Args:
            table: Table to query
            fields: Fields to query
            wheres: Where condition(s) for query
            max_rows: Maximum number of rows to return
            from_row: Starting row
            field_info: Metadata for each field in the table
            delimiter: Delimiter character used in data transfers

        Returns:
            Query results

        See Also:
            query
        """

        # Convert inputs to SAP Query format
        options = format_wheres(wheres, field_info)

        # Execute query
        # TODO: check if user has access to RFC "/BODS/RFC_READ_TABLE" or "/BODS/RFC_READ_TABLE2"
        RFCs = [
            "BBP_RFC__READ_TABLE",
            "RFC_READ_TABLE",
        ]
        rfc = RFCs[0]
        kwargs = dict(
            QUERY_TABLE=table,
            DELIMITER=delimiter,
            ROWCOUNT=max_rows,
            ROWSKIPS=from_row,
        )
        if fields:
            kwargs["FIELDS"] = fields
        if options:
            kwargs["OPTIONS"] = options
        result = self.call(rfc, **kwargs)
        logger.info(f"Ran {rfc}")

        # Parse results
        data = []
        for line in result["DATA"]:
            raw_data = line["WA"].strip().split(delimiter)
            data.append(
                [
                    field_value(value, result["FIELDS"][i]["TYPE"])
                    for i, value in enumerate(raw_data)
                ]
            )

        """data = [
            [
                field_value(value, result["FIELDS"][i]["TYPE"])
                for i, value in enumerate(line["WA"].strip().split(delimiter))
            ]
            for line in result["DATA"]
        ]"""
        """data = map(
            lambda line: [
                field_value(value, result["FIELDS"][i]["TYPE"])
                for i, value in enumerate(line["WA"].strip().split(delimiter))
            ],
            result["DATA"],
        )"""

        # Return results
        return data

    def number_entries(self, table: str) -> int:
        """Get number of entries in table

        Args:
            table: Table to query

        Returns:
            Number of entries in table
        """

        # Method 1
        # This actually reads the whole table into memory within SAP R/3!
        # result = self.call("RFC_GET_TABLE_ENTRIES", TABLE_NAME=table, MAX_ENTRIES=1)
        # logger.info("Ran RFC_GET_TABLE_ENTRIES")
        # entries = result["NUMBER_OF_ENTRIES"]

        # Method 2
        # Access often restricted to this RFC
        # result = self.call("EM_GET_NUMBER_OF_ENTRIES", IT_TABLES=[{"TABNAME": table}])
        # logger.info("Ran EM_GET_NUMBER_OF_ENTRIES")
        # entries = result["IT_TABLES"][0]["TABROWS"]

        # Method 3
        field_info = self.field_info(table)
        shortest_field = min(field_info, key=lambda field: field_info[field]["LENG"])
        data = self.query(table, shortest_field, max_rows=0)
        entries = len(data)

        return entries

    def table_description(self, table: str, language: str = "E") -> str:
        """Get table description

        Args:
            table: Table to query
            language: Language code

        Returns:
            Table description
        """

        fields = [
            "DDTEXT",  # short description
        ]
        wheres = [
            f"TABNAME = '{table}'",
            f"DDLANGUAGE = '{language}'",  # desired language
            "AS4LOCAL = 'A'",  # only active descriptions
        ]
        data = self.query("DD02T", fields, wheres, max_rows=0, field_info=False)
        if data:
            description = data[-1][0]
        else:
            description = None

        return description

    def find_tables_by_description(
        self, description: str, language: str = "E"
    ) -> list[list[str]]:
        """Find tables by description

        Args:
            description: Description to search for (% as wildcard)
            language: Language code

        Returns:
            List of tables and their descriptions
        """

        fields = [
            "TABNAME",  # table name
            "DDTEXT",  # short description
        ]
        wheres = [
            f"DDTEXT LIKE '{description}'",
            f"DDLANGUAGE = '{language}'",  # desired language
            "AS4LOCAL = 'A'",  # only active descriptions
        ]
        data = self.query("DD02T", fields, wheres, max_rows=0, field_info=False)

        return data

    def field_info(
        self, table: str, descriptions: bool = False, language: str = "E"
    ) -> dict[str, dict[str, str]]:
        """Get field information

        Args:
            table: Table to query
            descriptions: Include field text descriptions?
            language: Language code

        Returns:
            Field information
        """

        # Query basic field information
        fields = [
            "FIELDNAME",  # field name
            "KEYFLAG",  # key flag
            "ROLLNAME",  # data element (for getting field text from DD04T)
            "INTTYPE",  # ABAP data type
            "DATATYPE",  # data type
            "LENG",  # field length
            "POSITION",  # position in table
        ]
        wheres = [
            f"TABNAME = '{table}'",
            "AS4LOCAL = 'A'",  # only active descriptions
            "INTTYPE <> ''",  # has a data type
        ]
        data = self.query("DD03L", fields, wheres, max_rows=0, field_info=False)
        results = {
            row[0]: {fields[i]: row[1] for i in range(1, len(fields))} for row in data
        }

        if descriptions is True:
            # Query field text descriptions
            roll_index = fields.index("ROLLNAME")
            rollnames = [row[roll_index] for row in data]
            fields = [
                "ROLLNAME",  # data element
                "DDTEXT",  # short description
                "REPTEXT",  # heading
                "SCRTEXT_S",  # short field label
                "SCRTEXT_M",  # medium field label
                "SCRTEXT_L",  # long field label
            ]
            wheres = [
                ["ROLLNAME", "in", rollnames],
                f"DDLANGUAGE = '{language}'",  # desired language
                "AS4LOCAL = 'A'",  # only active descriptions
            ]
            field_info = {"ROLLNAME": {"LENG": 30, "INTTYPE": "C"}}
            data = self.query(
                "DD04T", fields, wheres, max_rows=0, field_info=field_info
            )
            roll_descriptions = {
                row[0]: {fields[i]: row[i] for i in range(1, len(fields))}
                for row in data
            }
            for field, info in results.items():
                roll = info["ROLLNAME"]
                if roll in roll_descriptions:
                    for key, value in roll_descriptions[roll].items():
                        results[field][key] = value
                del results[fields]["ROLLNAME"]

        return results

    def echo(self, message: str = "Hello world!"):
        """Echo a message

        Args:
            message: Message to echo

        Returns:
            Response from SAP R/3 System
        """
        response = self.call("STFC_CONNECTION", REQUTEXT=message)

        return response
